import random
from datetime import datetime

import numpy as np
import torch

try:
	import git
	has_git = True
except ModuleNotFoundError:
	has_git = False

def set_seeds(seed: int = 0) -> int:
	torch.manual_seed(seed)
	torch.cuda.manual_seed(seed)
	torch.cuda.manual_seed_all(seed)
	torch.backends.cudnn.deterministic = True
	torch.backends.cudnn.benchmark = False
	np.random.seed(seed)
	random.seed(seed)
	return seed

def get_commit() -> str:
	if has_git:
		repo = git.Repo(".")  # TODO: Search upwards in directories
		return str(repo.head.commit)
	return "Unknown (install GitPython to get this)"

def get_timestamp(for_file: bool = False) -> str:
	"""
	Returns a time stamp for current time either in datetime format or, if for_file, in YYYY-MM-DD_HH-MM-SS
	"""
	d_string = str(datetime.now())
	if for_file: d_string = "-".join(d_string.split(".")[0].split(":")).replace(" ", "_")
	return d_string

# To allow imports directly from utils #
# Currently to be placed lower because get_timestamp is needed by logger #
from .logger import *
from .parse import *
from .ticktock import *
from .datahandling import *



from time import perf_counter

from typing import List, Dict, Tuple

class TimeUnit:
	nanosecond  = ("ns",  1e9)
	microsecond = ("mus", 1e6)
	millisecond = ("ms",  1e3)
	second      = ("s",   1)
	minute      = ("min", 1/60)
	hour        = ("h",   1/3600)


class Profile:

	start: float

	def __init__(self, name: str, depth: int):
		self.hits: List[float] = []
		self.name = name
		self.depth = depth

	def get_hits(self):
		return self.hits

	def sum(self):
		""" Returns total runtime """
		return sum(self.get_hits())

	def mean(self):
		""" Returns mean runtime lengths """
		return self.sum() / len(self) if self.get_hits() else 0

	def std(self):
		""" Returns empirical standard deviation of runtime
		Be aware that this is highly sensitive to outliers and often a bad estimate """
		s = self.mean()
		return (1 / (len(self)+1) * sum(map(lambda x: (x-s)**2, self.get_hits()))) ** 0.5

	def remove_outliers(self, threshold=2):
		""" Remove all hits larger than threshold * average
		Returns number of removed outliers """
		mu = self.mean()
		original_length = len(self)
		self.hits = [x for x in self.hits if x <= threshold * mu]
		return original_length - len(self)

	def __str__(self):
		return self.name

	def __len__(self):
		return len(self.hits)

class TickTock:
	"""
	A taker that works like Matlab's Tic and Toc.
	Simples use case:
	```
	tt = TickTorck()
	tt.tick()
	<some task>
	time = tt.tock()
	```
	Profiling code sections is also supported
	```
	tt.profile(<profile name>)
	<some task>
	tt.end_profile()
	```
	"""
	_start = 0
	profiles: Dict[str, Profile] = {}
	_profile_stack: List[Profile] = list()

	def tick(self):
		self._start = perf_counter()
		return self._start

	def tock(self):
		end = perf_counter()
		return end - self._start

	def profile(self, name: str):
		if name not in self.profiles:
			self.profiles[name] = Profile(name, len(self._profile_stack))
		self._profile_stack.append(self.profiles[name])
		self.profiles[name].start = perf_counter()

	def end_profile(self, name: str=None):
		end = perf_counter()
		dt = end - self._profile_stack[-1].start
		self._profile_stack[-1].hits.append(dt)
		profile = self._profile_stack.pop()
		if name is not None:
			assert name == profile.name, f"Expected to pop profile '{profile.name}', received '{name}'"
		return dt

	def fuse(self, tt):
		"""Fuses a TickTock instance into self"""
		for profile in tt.profiles.values():
			if profile.name in self.profiles.keys():
				self.profiles[profile.name].hits += profile.hits
			else:
				self.profiles[profile.name] = profile

	def remove_outliers(self, threshold=2):
		# For all profiles, remove hits longer than threshold * average hit
		for profile in self.profiles.values():
			profile.remove_outliers(threshold)

	def reset(self):
		self.profiles = {}
		self._profile_depth = 0

	@staticmethod
	def thousand_seps(numstr: str or float or int) -> str:
		decs = str(numstr)
		rest = ""
		if "." in decs:
			rest = decs[decs.index("."):]
			decs = decs[:decs.index(".")]
		for i in range(len(decs)-3, 0, -3):
			decs = decs[:i] + "," + decs[i:]
		return decs + rest

	@classmethod
	def stringify_time(cls, dt: float, unit: Tuple[str, float]=TimeUnit.millisecond):
		str_ = f"{dt*unit[1]:.3f} {unit[0]}"
		return cls.thousand_seps(str_)

	def stringify_sections(self, unit: Tuple[str, float]=TimeUnit.second):
		# TODO: Less mess here
		# TODO: Keep track of children/parent profiles to ensure correct printing
		# Returns pretty sections
		strs = [["Execution times", "Total time", "Hits", "Avg. time"]]
		# std_strs = []
		for kw, v in self.profiles.items():
			# std = self.stringify_time(2*np.std(v["hits"]), "ms")
			# std_strs.append(std)
			strs.append([
				"- " * v.depth + kw,
				self.stringify_time(v.sum(), unit),
				self.thousand_seps(len(v)),
				self.stringify_time(v.mean(), TimeUnit.millisecond)
			])
		# longest_std = max(len(x) for x in std_strs)
		# std_strs = [" " * (longest_std-len(x)) + x for x in std_strs]
		# for i, str_ in enumerate(strs[1:]): str_[-1] += std_strs[i]
		for i in range(len(strs[0])):
			length = max(len(strs[j][i]) for j in range(len(strs)))
			for j in range(len(strs)):
				if i == 0:
					strs[j][i] += " " * (length - len(strs[j][i]))
				else:
					strs[j][i] = " " * (length - len(strs[j][i])) + strs[j][i]
		for i in range(len(strs)):
			strs[i] = " | ".join(strs[i])
		return "\n".join(strs)

	def __str__(self):
		return self.stringify_sections(TimeUnit.second)




import os
import json
from dataclasses import dataclass
from typing import List

import numpy as np

@dataclass
class DataStorage:
	"""The DataStorage class is an augmentation of the dataclass that incluces a standard way to save and load data.

	Currently works with:
		* Numpy arrays (numpy.ndarray)
		* Any type that can be saved with json.dump (e.g. convertible to str without data loss)

	The field `subfolder` chooses a defines a path to prepend all saved output.
	`json_name` chooses the name of the single json data file including all str data.

	Usage example
	```
	@dataclass
	class ResultData(DataStorage):
		shots: int
		goalscorers: list
		dists: np.ndarray

		subfolder = 'gamedata'
		json_name = 'game.json'

	rdata = ResultData(shots=1, goalscorers=["Max Fenger"], dists=np.ones(22)*10)
	rdata.save()
	# Now shots and goalscorers are saved in <pwd>/gamedata/game.json and dists in <pwd>/gamedata/dists.npy

	# In seperate script
	rdata = ResultData.load()
	print(rdata.goalscorers) #["Max Fenger"]
	```
	"""

	subfolder = ''
	json_name = 'data.json'

	def save(self, loc: str = '') -> List[str]:
		"""Saves all the fields of the instatiated data classes in npy's or json in <`loc`>/<`self.subfolder`>

		:param str loc: Save location to place the subfolder in
		:return: list (wtih length = # of saved files) of full save paths
		"""
		loc = self._get_loc(loc)
		if loc: os.makedirs(loc, exist_ok=True)

		# Split data by whether it should be saved .json and to .npy
		to_json, to_npy = dict(), dict()
		for key, data in self.__dict__.items():
			if isinstance(data, np.ndarray):
				to_npy.update({key: data})
			else:
				to_json.update({key: data})

		# Save data
		paths = [os.path.join(loc, self.json_name)]
		if to_json:
			with open(paths[0], "w", encoding="utf-8") as f:
				json.dump(to_json, f, indent=4)
		for key, arr in to_npy.items():
			paths.append(os.path.join(loc, f"{key}.npy"))
			np.save(paths[-1], arr)
		return paths

	@classmethod
	def load(cls, loc: str = ''):
		"""Instantiates the DataStorage-inherited class by loading all files saved by `save` of that same class.

		:param str loc: Save location used
		:return: An instance of this class with the content of the files
		"""
		loc = cls._get_loc(loc)
		any_json = False
		npys, non_numpy = dict(), dict()

		# Get .npy filenames and load these
		for key, field in cls.__dict__["__dataclass_fields__"].items():
			if field.type == np.ndarray:
				npys[key] = np.load(os.path.join(loc, f"{key}.npy"))
			else: any_json = True

		# Load .json if any non ndarray fields exist
		if any_json:
			with open(os.path.join(loc, cls.json_name), encoding="utf-8") as f:
				non_numpy = json.load(f)

		return cls(**non_numpy, **npys)

	@classmethod
	def _get_loc(cls, loc: str):
		return os.path.join(loc, cls.subfolder) if cls.subfolder else loc



from setuptools import setup, find_packages

with open('README.md') as readme_file:
	README = readme_file.read()

with open('HISTORY.md') as history_file:
	HISTORY = history_file.read()

# To update
# 1. Increment version below and document this in HISTORY.md (and possibly README.md)
# 2. Make master up to date with rebase
# 3. `rm -rf dist`
# 4. `python setup.py sdist bdist_wheel`
# 5. `twine upload dist/*`

setup_args = dict(
	name                          = 'pelutils',
	version                       = '0.1.2',
	description                   = 'Utility functions that are commmonly useful',
	long_description_content_type = "text/markdown",
	long_description              = README + '\n\n' + HISTORY,
	license                       = 'GPL-v2',
	packages                      = find_packages(),
	author                        = 'Søren Winkel Holm, Asger Laurits Schultz',
	author_email                  = 'swholm@protonmail.com',
	keywords                      = ['utility', 'logger', 'parser', 'profiling'],
	url                           = 'https://github.com/peleiden/pelutils',
	download_url                  = 'https://pypi.org/project/pelutils/'
)

install_requires = [
	'numpy',
	'torch',
]

if __name__ == '__main__':
	setup(**setup_args, install_requires=install_requires)



import os

from . import get_timestamp

class Unverbose:
	"""
	Used for disabling verbose logging in a code section
	Example:
	log = Logger(..., verbose=True)
	with unverbose:
		log("This will be logged")
		log.verbose("This will not be logged")
	"""
	allow_verbose = True

	def __enter__(self):
		self.allow_verbose = False

	def __exit__(self, *args):
		self.allow_verbose = True

unverbose = Unverbose()


class Logger:
	"""
	A simple logger which creates a log file and pushes strings both to stdout and the log file
	Sections, verbosity and error logging is supported
	"""
	_default_sep = "\n"

	def __init__(self, fpath: str, title: str, verbose=True):
		dirs = "/".join(fpath.split('/')[:-1])
		if not os.path.exists(dirs) and dirs:
			os.makedirs(dirs)

		self.fpath = fpath
		self._verbose = verbose

		with open(self.fpath, "w+", encoding="utf-8") as logfile:
			logfile.write("")

		self.log(title + "\n")

	def __call__(self, *tolog, with_timestamp=True, sep=None):
		self.log(*tolog, with_timestamp=with_timestamp, sep=sep)

	def log(self, *tolog, with_timestamp=True, sep=None):
		sep = sep or self._default_sep
		time = get_timestamp()
		with open(self.fpath, "a", encoding="utf-8") as logfile:
			tolog = sep.join([str(x) for x in tolog])
			spaces = len(time) * " "
			logs = tolog.split("\n")
			if with_timestamp and tolog:
				logs[0] = f"{time}\t{logs[0]}"
			else:
				logs[0] = f"{spaces}\t{logs[0]}"
			for i in range(1, len(logs)):
				logs[i] = f"{spaces}\t{logs[i]}"
				if logs[i].strip() == "":
					logs[i] = ""
			tolog = "\n".join(x.rstrip() for x in logs)
			logfile.write(tolog+"\n")
			print(tolog)

	def verbose(self, *tolog, with_timestamp=True):
		if self._verbose and unverbose.allow_verbose:
			self(*tolog, with_timestamp=with_timestamp)

	def is_verbose(self):
		return self._verbose and unverbose.allow_verbose

	def section(self, title=""):
		self.log()
		self.log(title)

	def throw(self, error: Exception, with_timestamp=True):
		self.log(error, with_timestamp=with_timestamp)
		raise error


class NullLogger(Logger):
	_verbose = False
	def __init__(self, *args, **kwargs):
		pass
	def log(self, *tolog, **kwargs):
		pass
	def section(self, title=""):
		pass



import os, sys
from argparse import ArgumentParser, RawTextHelpFormatter
from configparser import ConfigParser

from pprint import pformat


class Parser:
	"""
	Maintains multiple parsers to allow a workflow of mixing use of config files and CLI arguments.


	Can read from single .ini file and CLI arguments. CLI arguments overwrite settings in all experiments defined if ini file.
	In .ini file, defaults for all runs can be set in [DEFAULT] section and multiple runs can be added with their own section.

	The parser returns a list of dicts of the parsed settings for the experiments.


	Quick example:
	A file `main.py` could contain:
	```
	options = {
	location: {default: 'local_train', help: 'save_location', type: str},
	learning_rate: {default: 1.5e-3, help: 'Controls size of parameter update', type: float}
	}
	parser = Parser(options)
	experiments = parser.parse()

	```
	This could then by run by
	`python main.py --learning_rate 1e-5`
	or by
	`python main.py --config cfg.ini`
	where `cfg.ini` could contain

	```
	[DEFAULT]
	location = data/my_big_experiment
	[RUN1]
	learning_rate = 1e-4
	[RUN2]
	learning_rate = 1e-5
	```

	"""
	def __init__(self,
			options: dict,
			name: str = "Experiment",
			description: str = "Run experiments with these options",
			show_defaults: bool = True,
			description_last: bool = False,
		):
		"""
		Receives a dict of options.
		This should be a dict of dicts where each dict corresponds to an option with the key as the name.
		In the dict for each option, the key 'default' should store the default value and all other keys should correspond to
		kwargs in ArgumentParser.add_argument.
		"""
		self.options = options
		self.defaults = dict()
		self.save_location = ''
		self.name = name

		#Seperate parser for only receiving the config file
		self.config_receiver = ArgumentParser(add_help = False)
		self.config_receiver.add_argument('--config', help="Location of configuration file to use (if any). Config file should follow .ini format.", metavar='FILE')

		#Main parser for CLI arguments
		self.argparser = ArgumentParser(
			description=description,
			formatter_class=RawTextHelpFormatter,
			parents=[self.config_receiver]
		)
		if description_last:
			self.argparser.epilog = description
			self.argparser.description = None
		for argname, settings in self.options.items():
			self.defaults[argname] = settings.pop('default')

			if 'help' in settings and show_defaults:
				settings['help'] += f"\n  Default='{self.defaults[argname]}'"

			self.argparser.add_argument(f'--{argname}', **settings)

		self.configparser = ConfigParser()

	def parse(self, document = True) -> list:
		conf_arg, args = self.config_receiver.parse_known_args()

		experiments, with_config = self._read_config(conf_arg, args)

		if not experiments: #If configparser set nothing or only set defaults
			self.argparser.set_defaults(**self.defaults)
			args = self.argparser.parse_args(args)
			if args.location: self.save_location = args.location
			del args.config
			experiments.append({'name': self.name, **vars(args)})

		if document: self._document_settings(with_config)

		return experiments

	def _read_config(self, conf_arg, args) -> (list, bool):
		experiments = list()
		with_config = False

		if conf_arg.config:
			with_config = True
			if not self.configparser.read([conf_arg.config]):
				raise FileNotFoundError(f"Could not find config file {conf_arg.config}")

			# User set DEFAULT section should overwrite the defaults
			self.defaults = {**self.defaults, **dict(self.configparser.items("DEFAULT"))}

			# Each other section corresponds to an experiment
			for experiment_name in self.configparser.sections():
				options = {**self.defaults, **dict(self.configparser.items(experiment_name))}
				self.argparser.set_defaults(**options)
				exp_args = self.argparser.parse_args(args)

				#For multiple experiments, subfolders are nice
				if exp_args.location:
					if self.save_location and self.save_location != exp_args.location: raise ValueError("Multiple save locations are not supported")
					self.save_location = exp_args.location
					# Only give subfolder if there are indeed multiple runs
					if len(self.configparser.sections()) > 1: exp_args.location = f"{exp_args.location}/{experiment_name.lower()}"

				del exp_args.config
				experiments.append({'name': experiment_name, **vars(exp_args)})

		return experiments, with_config

	def _document_settings(self, with_config: bool):
		"""Saves all settings used for experiments for reproducability"""
		os.makedirs(self.save_location, exist_ok = True)

		with open(f"{self.save_location}/{self.name}_config.ini", 'w') as f:
			if with_config: self.configparser.write(f)
			f.write(f"\n# Run command\n# {' '.join(sys.argv)}\n")
			str_defaults = pformat(self.defaults).replace('\n', '\n# ')
			f.write(f"\n# Default configuration values at run\n# {str_defaults}")

